Require Import Coq.Lists.List.
Require Import Tactics.

Require Import Coq.Logic.FunctionalExtensionality.
Require Import Coq.Program.Basics.

Hint Unfold compose.

Import ListNotations.
Open Scope list_scope.
Print Scope list_scope.

Record Language := mkLang { t : Prop;
                            link : list t -> t -> t;
                            refines : t -> t -> Prop;
                            p_linkempty : (forall e : t, link nil e = e);
                            p_refineidentity : (forall e : t, refines e e);
                          }.
Set Printing Projections.

Hint Constructors Language.
Hint Resolve p_linkempty p_refineidentity.

Record LinkingSet {t : Prop} : Type := mkLink { phi : Prop;
                                             member : t * phi -> Prop;
                                           }.
Hint Constructors LinkingSet.

Definition CCC :=
  fun (S T hatS : Language)
    (compile : S.(t) -> T.(t))

    (* Cross-language linking & refinement *)
    (linkHat : list hatS.(t) -> S.(t) -> hatS.(t))
    (linkHat' : list S.(t) -> hatS.(t) -> hatS.(t))
    (refinesHatT : T.(t) -> hatS.(t) -> Prop)
    (refinesHatS : hatS.(t) -> S.(t) -> Prop)
    
    (L : @LinkingSet T.(t))
    (Lhat : @LinkingSet hatS.(t)) =>
    exists lift : T.(t) * L.(phi) -> hatS.(t) * Lhat.(phi),
    (*1*) (forall e : S.(t),
          forall etp : list (T.(t) * L.(phi)),
            List.Forall L.(member) etp ->
            refinesHatT (T.(link) (List.map fst etp) (compile e))
                        (linkHat (List.map (compose fst lift) etp) e))
    (*3*)  /\ (forall e' : S.(t), exists p' : L.(phi), L.(member) (compile e', p'))
    (*5*)  /\ (forall e' : S.(t), refinesHatS (linkHat nil e') e')
    (*6a*) /\ (forall etp' : list (T.(t) * L.(phi)),
                List.Forall L.(member) etp' ->
                forall es' : list S.(t),
                  ((forall c : T.(t), T.(refines) (T.(link) (List.map fst etp') c)
                                          (T.(link) (List.map compile es') c))
                   -> (forall c : S.(t), refinesHatS (linkHat (List.map (compose fst lift) etp') c)
                                               (S.(link) es' c))))
    (*6b*) /\ (forall ep' : T.(t) * L.(phi),
                 L.(member) ep' ->
                 forall e' : S.(t),
                   ((forall c : list T.(t), T.(refines) (T.(link) c (fst ep'))
                                                   (T.(link) c (compile e')))
                   -> (forall c : list S.(t), refinesHatS (linkHat' c ((compose fst lift) ep'))
                                                    (S.(link) c e')))).


               
Corollary CCCImpliesWholeProgram :
  forall S T hatS : Language,
  forall compile : S.(t) -> T.(t),
  forall refines : T.(t) -> S.(t) -> Prop,

  forall linkHat : list hatS.(t) -> S.(t) -> hatS.(t),
  forall linkHat' : list S.(t) -> hatS.(t) -> hatS.(t),
  forall refinesHatT : T.(t) -> hatS.(t) -> Prop,
  forall refinesHatS : hatS.(t) -> S.(t) -> Prop,

  forall L : @LinkingSet T.(t),
  forall Lhat : @LinkingSet hatS.(t),
  forall trans : (forall (t' : T.(t)) (h : hatS.(t)) (s : S.(t)), refinesHatT t' h -> refinesHatS h s -> refines t' s),
    CCC S T hatS compile linkHat linkHat' refinesHatT refinesHatS L Lhat ->
    forall e : S.(t),
      refines (compile e) e.
Proof.
  (* First, some preliminaries, to get our hypotheses and get names for parts of
  our CCC definition *)
  intros.
  unfold CCC in *.
  match goal with
  |[ H : exists _, _ |- _ ] => destruct H as (lift & CCCH)
  end.
  destruct CCCH as (C1 & C3 & C5 & rest).

  (* The main proof relies upon instantiating the condition 1 to the empty
  context *)
  assert (x1 : Forall L.(member) nil) by iauto.
  specialize (C1 e nil x1).
  simpl in C1.
  
  (* We also rely on that linking with empty contexts is an identity *)
  rewrite T.(p_linkempty) in C1.
  
  (* And proceed by composing two refinements (to be decided on) together *)
  eapply trans.
  (* The first is the refinement the compiled term and linkHat emptyHat e*)
  eapply C1.
  (* And the second is that linkHat with an emptyHat refines the term *)
  eapply C5.
Qed.


Corollary CCCImpliesSeparateCompilation :
  forall S T hatS : Language,
  forall compile : S.(t) -> T.(t),
  forall refines : T.(t) -> S.(t) -> Prop,

  forall linkHat : list hatS.(t) -> S.(t) -> hatS.(t),
  forall linkHat' : list S.(t) -> hatS.(t) -> hatS.(t),
  forall refinesHatT : T.(t) -> hatS.(t) -> Prop,
  forall refinesHatS : hatS.(t) -> S.(t) -> Prop,

  forall L : @LinkingSet T.(t),
  forall Lhat : @LinkingSet hatS.(t),

  forall trans : (forall (t' : T.(t)) (h : hatS.(t)) (s : S.(t)), refinesHatT t' h -> refinesHatS h s -> refines t' s),
    CCC S T hatS compile linkHat linkHat' refinesHatT refinesHatS L Lhat ->
    forall e1 e2 : S.(t),
      refines (T.(link) [compile e1] (compile e2))
              (S.(link) [e1] e2).
Proof.
  (* In this case, we again proceed with some preliminaries. *)
  intros.
  unfold CCC in *.
  match goal with
  |[ H : exists _, _ |- _ ] => destruct H as (lift & CCCH)
  end.
  destruct CCCH as (C1 & C3 & C5 & C6a & rest).

  (* We are going to later rely on the fact that we can link with the result of
  compiling e1, so we apply condition 3 to get that fact *)
  destruct (C3 e1) as (phi_e1 & link_e1).

  (* The main proof proceeds by applying condition 1 with (compile e1) as the
  target context and e2 as the term. *)
  assert (x1 : Forall L.(member) [(compile e1, phi_e1)]) by iauto.
  specialize (C1 e2 [(compile e1, phi_e1)] x1).

  (* We again will use the refinement transitivity to get the overall theorem.*)
  eapply trans.

  (* Where the first part is exactly the result of that use of C1 *)
  eapply C1.
  simpl.

  (* To get that linking lift(compile(e1)) with e2 refines linking e1 & e2, we
  instantiate condition 6 with (compile e1) as the target context and e1 as
  the source context. *)
  specialize (C6a [(compile e1, phi_e1)] x1 [e1]).
  (* This essentially does it, but we need a trivial identiity refinement as an
  obligation for condition 6. *)
  assert (refT : forall c : T.(t), T.(refines) (T.(link) [compile e1] c) (T.(link) [compile e1] c)) by iauto.
  (* And finally, we choose e2 as the arbitrary component being linked *)
  exact (C6a refT e2).
Qed.

Lemma map_map2 : forall (A B C D : Type) (f : A -> B) (g : B -> C) (h : C -> D) (l : list A),
    map (fun x : B => h (g x)) (map f l) = map (fun x : A => h (g (f x))) l.
Proof.
  intros.
  hint_rewrite map_map. iauto.
Qed.

Theorem PreciseBackTransitivity :
  (* NOTE(2018-02-05): In order to extract the intermediate lift function,
  as I have it set up, it seems like `I` needs to be in Prop. This is kind of
  unfortunate, but I don't see (right now) a simple way around this. *)
  forall S I T : Language,
  forall refinesTS : T.(t) -> S.(t) -> Prop,
  forall LS : @LinkingSet S.(t),
    
  forall compileSI : S.(t) -> I.(t),
  forall refinesIS : I.(t) -> S.(t) -> Prop,
  forall LI : @LinkingSet I.(t),
  forall CCC_SI : CCC S I S compileSI S.(link) S.(link) refinesIS S.(refines) LI LS,
             
  forall compileIT : I.(t) -> T.(t),
  forall refinesTI : T.(t) -> I.(t) -> Prop,
  forall LT : @LinkingSet T.(t),
  forall CCC_IT : CCC I T I compileIT I.(link) I.(link) refinesTI I.(refines) LT LI,

  (* NOTE(2019-02-01): We consider, as is the case for fully abstract
     compiler results we know of, cases where the only restrictions are based on
     what is safe to link, and thus the linking sets include _everything_. *)
  forall linkAllInI : (forall ip : I.(t) * LI.(phi), LI.(member) ip),
  forall linkAllInI' : (forall l : list (I.(t) * LI.(phi)), Forall LI.(member) l),
  forall linkAllInT : (forall tp : T.(t) * LT.(phi), LT.(member) tp),

  forall trans : (forall (t' : T.(t)) (i : I.(t)) (s : S.(t)), refinesTI t' i -> refinesIS i s -> refinesTS t' s),

    CCC S T S (fun s => compileIT (compileSI s))
        S.(link) S.(link) refinesTS S.(refines) LT LS.
Proof.
  intros.
  unfold CCC.
  exists (fun tp => (match CCC_IT return S.(t) * LS.(phi) with
            | ex_intro _ liftIT _ =>
              match CCC_SI return S.(t) * LS.(phi) with
                | ex_intro _ liftSI _ => 
                 (compose liftSI liftIT) tp
              end
             end)).


  (* C6a needs a little help because of the lists, but the map_map lemma, and a
  slightly specialized version of it (in the other direction), get it to go through
  as well. *)
  hint_rewrite map_map.
  hint_rewrite <- map_map2.

  unfold compose in *.
  
  (* At this point, we have 9 separate conditions to prove. *)
  repeat split;
    (* of which all but condition 1 follow easily from one of the two hypotheses. *)
    try solve [
          (unfold CCC in *;
          destruct CCC_SI as (liftSI & CCC_SI);
          destruct CCC_SI as (C1 & C3 & C5 & C6a & C6b);
          destruct CCC_IT as (liftIT & CCC_IT);
          destruct CCC_IT as (C1' & C3' & C5' & C6a' & C6b')); 
          intuition (eauto 6)
        ];
    [idtac].
  

  (* The interesting case is, of course, C1. *)

  intros.
  unfold CCC in CCC_IT.
  destruct CCC_IT as (liftIT & CCC_IT).
  destruct CCC_IT as (C1 & C3 & C5 & C6a & C6b).
  clear C3 C5 C6a C6b.

  (* At a high level, we instantiate the IT pass with (compileSI e) and the
  target components etp ... *)
  specialize (C1 (compileSI e) etp H).

  unfold CCC in CCC_SI.
  destruct CCC_SI as (liftSI & CCC_SI).
  destruct CCC_SI as (C1' & rest).
  clear rest.

  (* And the SI pass with e and the lifted versions of the target components etp. *)
  specialize (C1' e (map (fun t => liftIT t) etp)
                  (linkAllInI' (map (fun t : T.(t) * LT.(phi) => liftIT t) etp))).
  repeat (rewrite map_map in C1').
  simpl in C1'.

  (* At which point, we can simply compose the two together to get the result. *)
  eapply trans; iauto.
Qed.

Theorem CCCImpliesModularCorrectness :
  forall S T hatS : Language,
  forall compile1 : S.(t) -> T.(t),
  forall compile2 : S.(t) -> T.(t),
  forall equiv : (forall e : S.(t), forall c : T.(t),
                  T.(refines) (T.(link) [compile1 e] c)
                              (T.(link) [compile2 e] c)),
  forall refines : T.(t) -> S.(t) -> Prop,

  forall linkHat : list hatS.(t) -> S.(t) -> hatS.(t),
  forall linkHat' : list S.(t) -> hatS.(t) -> hatS.(t),
  forall refinesHatT : T.(t) -> hatS.(t) -> Prop,
  forall refinesHatS : hatS.(t) -> S.(t) -> Prop,

  forall L : @LinkingSet T.(t),
  forall Lhat : @LinkingSet hatS.(t),
  forall trans : (forall (t' : T.(t)) (h : hatS.(t)) (s : S.(t)), refinesHatT t' h -> refinesHatS h s -> refines t' s),
  forall CCC1 : CCC S T hatS compile1 linkHat linkHat' refinesHatT refinesHatS L Lhat,
  forall CCC2 : CCC S T hatS compile2 linkHat linkHat' refinesHatT refinesHatS L Lhat,
  forall e1 e2 : S.(t),
    refines (T.(link) [compile1 e1] (compile2 e2))
            (S.(link) [e1] e2).
Proof.
  intros.
  unfold CCC in *.
  
  destruct CCC1 as (liftSI & CCC1);
    destruct CCC1 as (C1 & C3 & C5 & C6a & C6b);
    destruct CCC2 as (liftIT & CCC2);
    destruct CCC2 as (C1' & C3' & C5' & C6a' & C6b').

  
  destruct (C3 e1) as (x & rest).

  assert (mems: Forall L.(member) [(compile1 e1, x)]) by iauto.
  specialize (C1' e2 [(compile1 e1, x)] mems).

  eapply trans.
  eapply C1'.

  eapply C6a'; iauto.

Qed.
